"use strict";
// Copyright 2023-2024 Design Liquido
// Copyright 2018 Johannes Wilm
// Copyright 2005 Google
//
// Original author: Steffen Meschkat <mesch@google.com>
//
// Miscellaneous utility and placeholder functions.
// Dummy implmentation for the logging functions. Replace by something
// useful when you want to debug.
Object.defineProperty(exports, "__esModule", { value: true });
exports.mapExec = mapExec;
exports.mapExpr = mapExpr;
exports.reverseInPlace = reverseInPlace;
// Applies the given function to each element of the array, preserving
// this, and passing the index.
function mapExec(array, func) {
    for (var i = 0; i < array.length; ++i) {
        func.call(this, array[i], i);
    }
}
// Returns an array that contains the return value of the given
// function applied to every element of the input array.
function mapExpr(array, func) {
    var ret = [];
    for (var i = 0; i < array.length; ++i) {
        ret.push(func(array[i]));
    }
    return ret;
}
/**
 * Reverses the given array in place.
 * @param array The array to be reversed.
 */
function reverseInPlace(array) {
    for (var i = 0; i < array.length / 2; ++i) {
        var h = array[i];
        var ii = array.length - i - 1;
        array[i] = array[ii];
        array[ii] = h;
    }
}
//# sourceMappingURL=util.js.map