/**
 * Our W3C DOM Node implementation. Note we call it XNode because we
 * can't define the identifier Node. We do this mostly for Opera,
 * where we can't reuse the HTML DOM for parsing our own XML, and for
 * Safari, where it is too expensive to have the template processor.
 */
export declare class XNode {
    id: number;
    childNodes: XNode[];
    nodeType: number;
    nodeName: string;
    nodeValue: any;
    firstChild: XNode;
    lastChild: XNode;
    nextSibling: XNode;
    previousSibling: XNode;
    siblingPosition: number;
    ownerDocument: any;
    namespaceUri: any;
    prefix: string;
    localName: string;
    parentNode: XNode;
    outputNode: XNode;
    transformedChildNodes: XNode[];
    transformedNodeType: any;
    transformedNodeName: string;
    transformedNodeValue: any;
    transformedFirstChild: XNode;
    transformedLastChild: XNode;
    transformedNextSibling: XNode;
    transformedPreviousSibling: XNode;
    transformedPrefix: any;
    transformedLocalName: string;
    transformedParentNode: XNode;
    visited: boolean;
    escape: boolean;
    static _unusedXNodes: any[];
    constructor(type: number, name: string, opt_value: any, opt_owner: any, opt_namespace?: any);
    /**
     * Node initialization. Called by the constructor and `recycle` method.
     * @param type The node type.
     * @param name The node name.
     * @param value The node value.
     * @param owner The node owner.
     * @param namespaceUri The node namespace.
     */
    init(type: number, name: string, value: string, owner: any, namespaceUri: any): void;
    protected qualifiedNameToParts(name: string): string[];
    protected domTraverseElements(node: XNode, opt_pre: Function, opt_post: any): boolean;
    static recycle(node: any): void;
    static create(type: any, name: string, value: any, owner: any, namespace?: any): XNode;
    static clone(node: XNode, newOwner: XNode): XNode;
    appendChild(node: XNode): void;
    appendTransformedChild(node: XNode): void;
    replaceChild(newNode: any, oldNode: any): void;
    insertBefore(newNode: any, oldNode: any): void;
    removeChild(node: XNode): void;
    hasAttributes(): boolean;
    setAttribute(name: string, value: any): void;
    setTransformedAttribute(name: string, value: any): void;
    setAttributeNS(namespace: any, name: any, value: any): void;
    getAttributeValue(name: string): any;
    getAttributeNS(namespace: any, localName: any): any;
    hasAttribute(name: string): boolean;
    hasAttributeNS(namespace: string, localName: string): boolean;
    removeAttribute(name: string): void;
    removeAttributeNS(namespace: string, localName: string): void;
    getElementsByTagName(name: string): any[];
    getElementsByTagNameNS(namespace: string, localName: string): any[];
    getElementById(id: any): any;
    getAncestorByLocalName(localName: string): XNode | undefined;
    getAncestorById(id: number): XNode | undefined;
}
