"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.copyArray = copyArray;
exports.copyArrayIgnoringAttributesWithoutValue = copyArrayIgnoringAttributesWithoutValue;
// Shallow-copies an array to the end of another array
// Basically Array.concat, but works with other non-array collections
function copyArray(dst, src) {
    if (!src)
        return;
    var dstLength = dst.length;
    for (var i = src.length - 1; i >= 0; --i) {
        dst[i + dstLength] = src[i];
    }
}
/**
 * This is an optimization for copying attribute lists in IE. IE includes many
 * extraneous properties in its DOM attribute lists, which take require
 * significant extra processing when evaluating attribute steps. With this
 * function, we ignore any such attributes that has an empty string value.
 */
function copyArrayIgnoringAttributesWithoutValue(dst, src) {
    if (!src)
        return;
    for (var i = src.length - 1; i >= 0; --i) {
        // this test will pass so long as the attribute has a non-empty string
        // value, even if that value is "false", "0", "undefined", etc.
        if (src[i].nodeValue) {
            dst.push(src[i]);
        }
    }
}
//# sourceMappingURL=common-function.js.map