"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.NodeTestName = void 0;
var boolean_value_1 = require("../values/boolean-value");
var NodeTestName = /** @class */ (function () {
    function NodeTestName(name) {
        this.name = name;
        if (name.indexOf(':') > 0) {
            var nameAndNamespacePrefix = name.split(':');
            this.namespacePrefix = nameAndNamespacePrefix[0];
            this.name = nameAndNamespacePrefix[1];
        }
        this.re = new RegExp("^".concat(name, "$"), 'i');
    }
    NodeTestName.prototype.evaluate = function (context) {
        var node = context.nodeList[context.position];
        if (this.namespacePrefix !== undefined) {
            var namespaceValue = context.knownNamespaces[this.namespacePrefix];
            if (namespaceValue !== node.namespaceUri) {
                return new boolean_value_1.BooleanValue(false);
            }
            if (context.caseInsensitive) {
                if (node.localName.length !== this.name.length)
                    return new boolean_value_1.BooleanValue(false);
                return new boolean_value_1.BooleanValue(this.re.test(node.localName));
            }
            return new boolean_value_1.BooleanValue(node.localName === this.name);
        }
        if (context.caseInsensitive) {
            if (node.nodeName.length !== this.name.length)
                return new boolean_value_1.BooleanValue(false);
            return new boolean_value_1.BooleanValue(this.re.test(node.nodeName));
        }
        return new boolean_value_1.BooleanValue(node.nodeName === this.name);
    };
    return NodeTestName;
}());
exports.NodeTestName = NodeTestName;
//# sourceMappingURL=node-test-name.js.map